#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
### PYTHON. APPRENTISSAGE ACTIF (JP. ROY, EDITIONS ELLIPSES, 2020) ###
######################################################################

"""
exo_aut1d.py (11.86)
Automates cellulaire 1-D : les règles de Wolfram
"""
 
# ----------- le modèle de calcul ------------------------------------

# Une population de cellules vivantes est une liste [v,...] de valeurs
# 0 (morte) ou 1 (vivante). Une "règle d'évolution" est un nombre entier
# r dans [0,255]. On interprète les chiffres binaires de r comme les
# valeurs de la cellule à la génération suivante suivant ses deux voisins.
# On ne recolle pas les bords : les cellules aux extrémités ont un
# voisin toujours mort, que l'on place en "sentinelle". La portion de
# travail ne comprend pas les cellules aux extrémités. 

from random import randint

N=300                           # pour une fenêtre tortue 600 x 600

def population(strf):
    '''strf est une chaîne de 0,1 ou une fonction 0..n-1 -> {0,1}.
    Le résultat est une liste de longueur n+2 formée de 0,1. Si strf
    est une chaîne, elle sera tronquée si trop longue, ou placée au
    milieu d'une chaîne de longueur long et complétée par des 0.'''
    if type(strf) == str:       # une suite uniquement de '0' et '1' ?
        if any(map(lambda c: c not in '01',strf)):    # on le vérifie
            raise ValueError('On attend une chaîne de 0,1')
        k = len(strf)
        if k > N:
            raise ValueError("Chaîne d'initialisation trop longue.")
        elif k < N:
            #print("Chaîne d'initialisation trop courte : augmentée !")
            p = (N - k) // 2
            res = p*[0] + list(map(int,strf)) + (N-k-p)*[0]
        else: 
            res = list(map(int,strf))
    elif callable(strf):       # strf est une fonction
        res = [int(strf(i)) for i in range(N)]
    else:
        raise(ValueError('On attendait une chaîne de 0,1 ou une fonction.'))
    assert len(res) == N
    res.append(0)              # 0 en sentinelle aux extrémités
    return [0] + res           # on rend une liste de longueur N+2

def regle(n):
    '''Prend un numéro entier n de règle entre 0 et 255, et retourne
    une liste de longueur 8 contenant les chiffres de r en binaire.'''
    assert 0 <= n <= 255
    res = [int(i) for i in bin(n)[2:]]    # conversion en string binaire
    res = (8 - len(res))*[0] + res        # puis en liste de 0,1 de longueur 8
    return res

def afficher_regle(n):
    print('-------------------------------')
    print('Règle',n,':')
    print('000 001 010 011 100 101 110 111',end='\n ')
    for i in range(8):
        print(n % 2,end='   ')
        n //= 2
    print('\n-------------------------------')
    
def evolution(L,Lr):
    '''Retourne la génération suivante de la population L avec la
    règle-liste Lr == regle(n).'''
    LS = len(L)*[0]
    for i in range(1,len(LS)-1):     # touche pas aux sentinelles !
        v = L[i-1:i+2]               # moi entre mes deux voisins, ex: [1,1,0]
        k = int(str(v[0]) + str(v[1]) + str(v[2]),2)    # ex: k=5
        # extraction du bit numéro 7-k de r (bit 0 à droite)
        LS[i] = Lr[7-k]
    return LS

#------------- la vue en mode texte ------------------------------------

def afficher_texte(L,n,nbgen=20):
    '''Simple outil pour la mise au point. Affiche en mode texte la 
    population L de cellules pendant n générations. L est ici une liste
    de 0 et 1 avec un 0 à chaque extrémité, et de longueur quelconque'''
    print('L =',L)
    Lr = regle(n)
    afficher_regle(n)
    for i in range(1,nbgen+1):
        print(i,end='\t')
        for i in range(1,len(L)-1):
            print(L[i],end=' ')
        L = evolution(L,Lr)
        print()

#------------- la vue en mode graphique (tortue) ----------------------

from turtle_init import *         # cf ci-dessous
# init(x=0,y=0,cap=0)
# afficher(msg,x,y)
# sleep(dt)            en secondes flottantes
# deg_rad(a_degres)    conversion degrés -> radians
# rot(M,K,a_degres)    K = centre de rotation
# segment(M1,M2)       le tracé
# pi, sqrt, sin, cos, randint   pour les calculs fréquents

def afficher_graphique(strf,n,nbgen=N//2):
    '''Affiche N générations à partir de la population L avec la
    règle numéro r.'''
    afficher_regle(n)                  # au toplevel
    reset() ; ht()
    tracer(False)
    afficher('Règle {}'.format(n),-260,300)    # importé de turtle_init
    up()
    r = regle(n)                       # int --> list
    L = population(strf)
    for y in range(nbgen):             # ligne y
        for x in range(1,len(L)-1):    # colonne x, on n'affiche pas les sentinelles
            goto(-300+2*x,300-2*y)
            if L[x] == 1: dot(2)       # le minimum...
        L = evolution(L,r)             # génération suivante
    tracer(True)

#-------------- tests -------------------------------------------------
    
if __name__ == '__main__':
    # triangle de Sierpinsky (ou triangle de Pascal modulo 2)
    print("Test en cours de l'affichage en mode graphique.")
    NUM = 90                                                # numéro de la règle
    afficher_graphique('1',NUM)
    getcanvas().postscript(file='regle{}.ps'.format(NUM))   # image Postscript sur disque
    print('Fichier "regle{}.ps" produit.\n'.format(NUM))
    
    print("Tests de l'affichage en mode texte.")
    afficher_texte([0,0,0,0,0,1,0,1,1,0,0,0,0,0],30,nbgen=4)
    print()
    afficher_texte(10*[0]+[1]+10*[0],90)
    
    mainloop()
    
    # COMPLEMENTS EN LIGNE
    from webbrowser import open as browse
#    browse('https://www.wolframscience.com')
#    browse('https://www.rennard.org/alife/french/ac.pdf')
#    browse('http://mathworld.wolfram.com/ElementaryCellularAutomaton.html')
#    browse('http://musiquealgorithmique.fr/automates-cellulaires/')
#    browse('https://fr.wikipedia.org/wiki/Automate_cellulaire')


