#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
### PYTHON. APPRENTISSAGE ACTIF (JP. ROY, EDITIONS ELLIPSES, 2020) ###
######################################################################

"""
exo_diff.py (11.107)
Dérivation (mécanique) d'un arbre binaire par-rapport à une variable.
On utilise l'embryon de simplificateur.
"""

from exo_simplif import simplif
from arbres2 import *

def diff(a,var):
    '''Retourne l'arbre derive de l'arbre a par-rapport a la variable var.'''
    if est_feuille(a): return 1 if a == var else 0
    r = racine(a)
    if r in ('+','-'): return diff_add(a,var)
    if r == '*': return diff_mul(a,var)
    if r == '/': return diff_div(a,var)
    raise ValueError('Operateur inconnu {}'.format(r))

def diff_add(a,var):       # a == ['+',u,v]
    dg = diff(fg(a),var)
    dd = diff(fd(a),var)
    return simplif(arbre(racine(a),dg,dd))

def diff_mul(a,var):       # a == ['*',u,v]
    dg = arbre('*',diff(fg(a),var),fd(a))   # u'*v
    dd = arbre('*',fg(a),diff(fd(a),var))   # u*v'
    return simplif(arbre('+',dg,dd))

def diff_div(a,var):       # a == ['/',a1,a2]
    dg = arbre('-',arbre('*',diff(fg(a),var),fd(a)),arbre('*',fg(a),diff(fd(a),var)))
    dd = arbre('*',fd(a),fd(a))     # hélas, pas d'opérateur **
    return simplif(arbre('/',dg,dd))

from exo_arbrebin import valeur # valeur d'un arbre algébrique
from math import factorial
from fractions import Fraction

def valeur(a,mem):
    '''Retourne la valeur de l'arbre binaire algebrique a dans une
    memoire mem. L'operateur / sera la division flottante.'''
    if est_feuille(a) :
        if type(a) in (int,float,Fraction) : return Fraction(a)
        try : return Fraction(mem[a])
        except KeyError : raise ValueError('Variable indéfinie {}'.format(a))
    vg = valeur(fg(a),mem)
    vd = valeur(fd(a),mem)
    op = racine(a)
    if op == '+' : return vg + vd
    if op == '-' : return vg - vd
    if op == '*' : return vg * vd
    if op == '/' : return Fraction(vg,vd)
    raise ValueError('Opérateur inconnu : {}'.format(op))

def taylor(a,x='x',x0=0,n=4):  # demanderait un véritable simplificateur...
    '''Retourne la liste des coeffs du developpement de Taylor de l'arbre a
    a l'ordre n au point x0 et par-rapport a la variable var.'''
    mem = {x:Fraction(x0)}     # calculs rationnels si possible
    res = [valeur(a,mem)]
    for i in range(1,n+1) :
        a = diff(a,x)
        va = valeur(a,mem)/factorial(i)    # la fonction valeur renvoie une Fraction
        res.append(va)
    return res

if __name__ == '__main__':
    print('# d/dx (3x**2+x*y)')
    print(">>> diff(['+',['*',3,['*','x','x']],['*','x','y']],'x')")
    print('-->',diff(['+',['*',3,['*','x','x']],['*','x','y']],'x'))
    print()
    print('# d/dy (3x**2+x*y)')
    print(">>> diff(['+',['*',3,['*','x','x']],['*','x','y']],'y')")
    print('-->',diff(['+',['*',3,['*','x','x']],['*','x','y']],'y'))
    print()
    print('# d/dx (3*x*y-x/y)')
    print(">>> diff(['-',['*',3,['*','x','y']],['/','x','y']],'x')")
    print('-->',diff(['-',['*',3,['*','x','y']],['/','x','y']],'x'))
    print()
    print('# d/dy (3*x/y)')
    print(">>> diff(['/',['*',3,'x'],'y'],'y')")
    print('-->',diff(['/',['*',3,'x'],'y'],'y'))
    print()
    print("# Développement de Taylor exact de 1/(1+2*x) en x=2 à l'ordre 4 :")
    a = ['/',1,['+',1,['*',2,'x']]]
    print(taylor(a,x0=2,n=4))
    # mais mortel dès que l'ordre monte, car la taille des dérivées
    # explose à cause d'un simplificateur (beaucoup) trop naïf :-(
    print()
    print('Vérification  de Taylor avec sympy :')
    from sympy import var, series
    var('x')        # variable symbolique (demandez la valeur de x)
    print('>>> series(1/(1+2*x),x,x0=2,n=4)')
    print('-->',series(1/(1+2*x),x,x0=2,n=5))
    print()
    # OK même si le développement est moins bien présenté par sympy...
    print('Vérification de Taylor avec Mathematica dans le navigateur Web.')
    from webbrowser import open as browse
    browse('https://www.wolframalpha.com/input/?i=series+expansion+of+1%2F(1%2B2x)+for+x%3D2+with+order+5')

    # COMPLEMENTS EN LIGNE
    #browse('https://www.youtube.com/watch?v=QlRB2k9i4gc')
    #browse('https://maths-with-python.readthedocs.io/en/latest/07-sympy.html')
    
    # le lien suivant est à la pointe, sur la "différentiation automatique" (AD)
    # avec l'algorithme de rétro-propagation utilisé en "machine learning"
    # (cf livre §11.134, exos sur les réseaux de neurones). Mais attention,
    # niveau avancé :
    #browse('http://jmlr.org/papers/volume18/17-468/17-468.pdf')







    

