#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
### PYTHON. APPRENTISSAGE ACTIF (JP. ROY, EDITIONS ELLIPSES, 2020) ###
######################################################################

"""
exo_excel.py (11.120)
Lecture et écriture dans une feuille Excel.
Calcul par Excel d'une droite de régression linéaire.
"""

# -----------------------------------------------------------------
# ETAPE 1 : simulation en Python de données bruitées.
# On génère des couples (x,y) situés autour de la droite y = 2x+10

# Utilitaire de réduction à 2 chiffres décimaux max. pour Excel
def reduc(x):       # retourne x comme flottant avec 2 chiffres décimaux
    return float('{:.2f}'.format(x))

# La même fonction avec une "f-string" dont le livre ne parle pas, apparue
# avec Python 3.6. Grosso modo :    '{:.2f}'.format(x) == f'{x:.2f}'
def reduc(x):
    return float(f'{x:.2f}')                 # hum, cool ?

from random import uniform            # tirage de flottants aléatoires
N = 20                                       # nombre de données
L1 = list(range(N))                          # abscisses 0,1,2,...,N-1
L2 = [reduc(2*i+20*uniform(0,1)) for i in range(N)]   # ordonnées bruitées

# -----------------------------------------------------------------
# ETAPE 2 : calcul de la droite de régression linéaire y = ax+b

# On peut faire les calculs en Python pur avec les formules
# bien connues de la régression en stats, ou mieux directement avec SciPy :
import numpy as np
from scipy import stats
(a, b, r, _, _) = stats.linregress(L1,L2)
# a = pente, b = ordonnée à l'origine, r = coeff. de corrélation
print('Régression avec scipy :\na={}, b={}, r={}'.format(a,b,r))

# -----------------------------------------------------------------
# ETAPE 3 : placement de tout ça dans une feuille vierge d'un classeur Excel.

from openpyxl import Workbook       # book = classeur, sheet = feuille

wb = Workbook()                     # un classeur vierge (workbook)
ws = wb.active                      # la feuille active (worksheet)
ws.title = 'regression'             # nom de la feuille active

# installation des données sur les deux premières lignes
ws.append(['x']  + L1)              # les abscisses
ws.append(['y']  + L2)              # les ordonnées
# installation des ordonnées de la droite de régression sur la troisième ligne
Lr = [reduc(a*x+b) for x in L1]    # ordonnées sur la droite de régression
ws.append(['yr'] + Lr)

# Installation de cellules individuelles
ws['A5'] = 'DROITE DE REGRESSION LINEAIRE y = ax+b'
ws['A6'] = 'a =' ; ws['B6'] = reduc(a)     # proche de 2 ?
ws['D6'] = 'b =' ; ws['E6'] = reduc(b)     # proche de 10 ?
ws['G6'] = 'r =' ; ws['H6'] = reduc(r)     # proche de 1 ?

# Style : toutes les cellules en "Courier 14"
from openpyxl.styles import NamedStyle, Font
print('Zone stylée de la feuille de calcul',ws.calculate_dimension())
mon_style = NamedStyle('mon_style')
mon_style.font = Font(name='Courier',bold=False,italic=False,size=14)
for row in ws.rows:        # pour toutes les lignes row utilisées de ws
    for cell in row:       # et toutes les cellules de la ligne row
        cell.style = mon_style
        cell.number_format = '0.00'

# on pouvait aussi rédiger la boucle en utilisant l'accès à une cellule
# donnée sous la forme ws.cell(column=8,row=5).

# -----------------------------------------------------------------
# ETAPE 4 : construction du graphique Excel : nuage de points + droite
# Donc deux séries numériques y et yr basées sur la série des x

from openpyxl.chart import ScatterChart, Reference, Series

chart = ScatterChart()
chart.title = "Droite de régression linéaire"
chart.style = 13
chart.x_axis.title = 'x'
chart.y_axis.title = 'y'
# une première série des (x,y) dans L1 et L2
x_values = Reference(ws, min_col=2, min_row=1, max_col=21)  # à partir de 1
y_values = Reference(ws, min_col=2, min_row=2, max_col=21)
serie1 = Series(y_values, x_values,title_from_data=True)
serie1.marker.symbol = 'circle'                             # points-disques
serie1.marker.graphicalProperties.solidFill = "FF0000"      # rouges à l'intérieur
serie1.marker.graphicalProperties.line.solidFill = "FF0000" # avec bord rouge
serie1.graphicalProperties.line.solidFill = "000000"        # reliés en noir
serie1.graphicalProperties.line.width = 100                 # en EMU
chart.series.append(serie1)                                 # chart est le graphique
# puis on superpose une seconde série des (x,yr) issue de la droite de régression
yr_values = Reference(ws, min_col=2, min_row=3, max_col=21)
serie2 = Series(yr_values, x_values,title_from_data=True)
serie2.graphicalProperties.line.solidFill = "0000FF"    # ligne bleue
serie2.graphicalProperties.line.width = 40000           # en EMU (cf page web plus bas)
chart.series.append(serie2)
# et on place le graphique obtenu
ws.add_chart(chart,'B8')                                # B8 en coin haut gauche

# ------- et les FORMULES ? ---------
# Faiblesse de openpyxl avec les formules. On peut insérer une formule comme :
ws['A24'] = '=INVERSEMAT(B1:D3)'     # inverse de la matrice en B1:D3
ws['G24'] = '=SOMME(B1:G1)'          # somme des valeurs de B1:G1
# mais le calcul n'est pas fait automatiquement, il faut intervenir à la main
# dans la feuille de calcul :-(
# - Sélectionnez la zone A24:C26 qui contiendra la matrice inverse, puis tapez
#   Ctrl-U suivi de Cmd-Ret (sur Mac).
# - Sélectionnez la cellule G24 puis Ctrl-U suivi de Cmd-Ret (sur Mac).
# Ctrl-U fait apparaître la formule, et Cmd-Ret provoque le calcul.

# ETAPE 5 : sauvegarde du classeur Excel au format 2010 (xlsx/xlsm/xltx/xltm)

wb.save(filename='regression.xlsx')                   # et zou !

# COMPLEMENTS EN LIGNE
from webbrowser import open as browse
#browse('https://realpython.com/python-f-strings/')
#browse('https://startbigthinksmall.wordpress.com/2010/01/04/points-inches-and-emus-measuring-units-in-office-open-xml/')
#browse('https://www.rickyadams.com/wp/index.php/2017/04/25/using-python-with-excel-spreadsheets/')
#browse('https://openpyxl.readthedocs.io/en/stable/index.html')



