#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
### PYTHON. APPRENTISSAGE ACTIF (JP. ROY, EDITIONS ELLIPSES, 2020) ###
######################################################################

"""
exo_genetic.py (11.129)
Ajustement d'une courbe par programmation génétique.
On utilise numpy pour le graphisme seulement...
"""

# Préparation des données. La courbe que l'on recherche sur [0,8],
# l'inconnue, est paramétrée par a,b,c,d. Il faut trouver les valeurs
# de a,b,c,d pour que la courbe passe au plus près des données
# expérimentales.

from math import pi, cos, exp

def osc(t,a,b,c,d):           # but : trouver les valeurs de a,b,c,d
    return a * exp(b*t) * cos(c*t + d)   # une exponentielle amortie (b < 0)

# Les données (t,g(t)) pour t dans [0,8] où g est la solution exacte (chut !)

g = lambda t: osc(t,2.5,-0.4,8,0)        # g(t) == 2.5*exp(-0.4*t)*cos(8*t)
Lt = [t/100 for t in range(0,801,20)]    # 41 nombres flottants sur [0,8]
DATA = {t:g(t) for t in Lt}              # nuage de 41 points
print('DATA =',DATA)
print()

# Les individus

from random import uniform

class Individu:
    '''Un objet de classe Individu représente une fonction :
    t -> a * exp(b * t) * cos(c*t + d). Les "gènes" sont a,b,c,d.'''
    
    def __init__(self): 
        '''Construit un Individu dont les attributs sont a,b,c,d
        choisis "au hasard", et dont l'aptitude initiale est très faible,
        i.e. très loin de la solution.'''
        self.a = uniform(-10,10)
        self.b = uniform(-10,10)          # l'oscillation est amortie dans notre exemple
        self.c = uniform(-10,10)          # paramètre de fréquence
        self.d = uniform(-10,10)          # paramètre de phase
        # NB. Remplacez (-10,10) par (0,10). Trouvez-vous la même solution ?
        self.fitness = float('inf')     # distance initiale à la solution

    def valeur(self,t):
        '''Calcule la valeur de l'individu au point t.'''
        return osc(t,self.a,self.b,self.c,self.d)
    
    def evaluer(self):   # éloignement de self à la solution avec des moindres carrés, à minimiser !
        '''Retourne la valeur de la fonction objective, le "fitness", ici la somme des carres
        des écarts aux données situées dans DATA.'''
        self.fitness = sum((self.valeur(t) - DATA[t])**2 for t in DATA.keys())

    def mutation(self,amplitude):
        '''Produit une mutation d'une certaine amplitude autour des valeurs
        courantes des gènes de l'individu courant.'''
        self.a += uniform(-amplitude,amplitude)
        self.b += uniform(-amplitude,amplitude)
        self.c += uniform(-amplitude,amplitude)
        self.d += uniform(-amplitude,amplitude)

# Les populations d'individus

from copy import deepcopy

class Population:
    '''Une population est un groupement d'individus destinés à évoluer
    suivant les lois de la génétique pour s'approcher de la solution
    d'un problème, ici un ajustement de courbe.'''
    def __init__(self,N=100,amp=0.1):       # N individus
        self.N = N
        self.L = [Individu() for i in range(N)]
        for x in self.L: x.evaluer()
        self.trier()                # la population est triée par valeur de fitness croissante
        self.amp = amp              # amplitude max. d'une mutation d'un gène

    def evaluer(self):
        '''Demande à chaque individu de la population de procéder à son évaluation et retourne
        l'évaluation totale de la population, qu'il s'agit de minimiser.'''
        fitness_total = 0
        for x in self.L:
            x.evaluer()             # met à jour la valeur de fitness de l'individu x
            fitness_total += x.fitness
        return fitness_total

    def trier(self):
        '''Range les individus par valeur de fitness croissante.
        Donc les meilleurs sont en tête.'''
        self.L = sorted(self.L, key = lambda x: x.fitness)
    
    def ajuste(self):
        '''Essaye d'ameliorer la population par selection et mutation des individus.'''
        # La population est déjà triée !
        # Les 20% meilleurs, on n'y touche pas pour la population suivante, mais
        # pour chacun d'entre eux, on construit 4 clones avec une petite mutation
        # génétique pour chacun. Pour assurer la variété génétique...
        meilleurs = self.L[:self.N//5]              # les 20% meilleurs en tête
        nouvelle_population = deepcopy(meilleurs)   # clonage en profondeur des meilleurs
        for x in meilleurs:                         # et pour chacun des meilleurs :
            for i in range(4):                      # création de quatre
                y = deepcopy(x)                     # clones
                y.mutation(self.amp)                # mutants (d'où la deepcopy !)
                nouvelle_population.append(y)
        self.L = nouvelle_population      # on a remplacé 80% de l'ancienne population...
        self.evaluer()
        self.trier()

    def evolue(self,nbgen,verbeux=False):
        '''Demande à la population courante de s'ajuster nbgen fois.'''
        for i in range(nbgen):
            self.ajuste()
            if verbeux:
                meilleur = self.L[0]
                print('Génération {}, meilleur fitness : {}'.format(i,meilleur.fitness))

# ------- le programme de sélection génétique --------------------

popu = Population(N=200)                # avec 200 individus aléatoires
popu.evolue(nbgen=200,verbeux=False)    # sur 300 générations !
# on pourrait laisser évoluer jusqu'à ce que le fitness du meilleur soit assez petit...
meilleur = popu.L[0]                    # le meilleur individu
print('Le meilleur individu a pour fitness {:.5f}'.format(meilleur.fitness))
print('Solution exacte  :\n  ',[2.5, -0.4, 8, 0])
[a,b,c,d] = [meilleur.a,meilleur.b,meilleur.c,meilleur.d]
print('Gènes du meilleur individu :\n  ',[a,b,c,d])       # d est une phase...
print()

# -------- comparaison avec les données exactes ------------------

f = lambda t : osc(t,a,b,c,d)
print('Comparaison avec les données exactes :')
print(' t\tDATA[t]\t\tsol_approx(t)')
for t in DATA.keys():
    print('{}\t{:.2f}\t\t{:.2f}'.format(t,DATA[t],f(t)))
print()

# ---- affichage des points de DATA et de la courbe trouvée ------
# conversion des données vers des tableaux numpy : list -> array

import numpy as np
import matplotlib.pyplot as plt

tDATA = np.asarray(Lt)
yDATA_exact = np.asarray([DATA[t] for t in Lt])
plt.plot(tDATA,yDATA_exact,'k.')        # le nuage de points DATA en noir

tDATA_ajust = np.linspace(0,8,200)      # on veut une courbe lisse : 200 points
uf = np.frompyfunc(f,1,1)               # f devient une fonction "universelle" de numpy (ufunc)
yDATA_ajust = uf(tDATA_ajust)           # composante à composante puisque uf est une ufunc
plt.plot(tDATA_ajust,yDATA_ajust,'r-')  # la courbe approchée en rouge
plt.show()                              # et zou.
print()

# COMPLEMENTS EN LIGNE
from webbrowser import open as browse
browse('https://fr.wikipedia.org/wiki/Programmation_génétique')
browse('https://www-lisic.univ-littoral.fr/~fonlupt/Recherche/PG/prog-genetique.html')

# --------- lecture "en passant" totalement hors-sujet : Maxima -----------------------
# Maxima est un patriarche encore vivant parmi les systèmes de calcul formel. Richard Fateman
# de Berkeley, a beaucoup travaillé sur ce logiciel et produit nombre de papiers intéressants.
# Maxima est écrit en Lisp et se programme dans une syntaxe spéciale mais assez proche
# d'un Lisp sans parenthèses... SAGE comporte certaines librairies empruntées à Maxima.
# https://wxmaxima-developers.github.io/wxmaxima/
# http://hyperpolyglot.org/computer-algebra
# http://michel.gosse.free.fr
print('Pour coller dans wxMaxima :')
print('plot2d({:.3f}*%e^({:.3f}*t)*cos({:.3f}*t+({:.3f})),[t,0,8]);'.format(a,b,c,d))


        
