#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
### PYTHON. APPRENTISSAGE ACTIF (JP. ROY, EDITIONS ELLIPSES, 2020) ###
######################################################################

"""
exo_pgcd.py (11.21)
Le Plus Grand Common Diviseur, une des gloires d'Euclide
"""

def pgcd(a,b):               # a et b entiers > 0, RECURRENCE
    if b == 0: return a      # on utilise directement le théorème d'Euclide
    return pgcd(b,a % b)     # une merveille de simplicité !

# L'algorithme est rapide, la récurrence est efficace. Le "théorème de Lamé"
# (1845) affirme que le nombre d'application de la règle d'Euclide pour 
# calculer le pgcd(a,b) est de l'ordre du Min(log(a),log(b)). Voir 
# [ABE] § 1.2.5 dans la bibliographie pour une preuve utilisant les nombres
# de Fibonacci. La croissance du calcul étant logarithmique, il ne faut pas
# s'étonner que l'algorithme soit rapide et fonctionne avec de grands entiers...

def pgcd(a,b):               # a et b entiers > 0, ITERATION
    while b > 0:
        (a,b) = (b,a % b)    # différent de:    a = b ; b = a % b
    return a

# IMHO, je trouve la récurrence plus élégante...

from math import gcd        # il existe aussi lcm pour le ppcm, moins utilisée
                            # et de plus : gcd(a,b) * lcm(a,b) == a * b

def premiers_entre_eux(a,b):          # a et b entiers > 0
    return gcd(a,b) == 1

# La probabilité théorique que deux entiers au hasard soient premiers entre
# eux est égale à 6/pi**2. Vérifions-le :

def proba_premiers_entre_eux():
    nb_succes = 0
    for i in range(100000):
        a = randint(2,10000000000)
        b = randint(2,10000000000)
        if gcd(a,b) == 1: nb_succes += 1
    return nb_succes/100000.0

# DECOMPOSITION DE BEZOUT DU PGCD. Non unique...
# - Si b=0, le triplet (a,1,0) fonctionne : a = a*1 + 0*b.
# - Sinon, supposons par hyp. de récurrence que l'on ait la solution (g,u,v)
#   de bezout(b,a % b) en suivant l'idée d'Euclide. Alors g est aussi le pgcd
#   de a et b : c'est le théorème d'Euclide. Et l'on a : g = b * u + (a%b) * v.
#   Posons q = a // b et r = a % b. On a donc g = b * u + r * v.
#   Mais a = b * q + r, d'où g = b * u + (a - b*q) * v = a * v + b * (u - q*v)
#   D'où un triplet de Bezout (g,v,u-q*v) pour a et b. CQFD
#   Cette preuve par récurrence se traduit en un algorithme bien plus lisible
#   que les algorithmes itératifs usuels... 

def bezout(a,b):                 # a et b entiers > 0
    if b == 0: return (a,1,0)
    (q,r) = (a // b, a % b)      # division euclidienne
    (g,u,v) = bezout(b,r)        # hyp. de récurrence
    return (g,v,u-q*v)           # et zou.

def aff_bezout(a,b):             # aucun résultat
    (g,u,v) = bezout(a,b)
    print('{} = {} * ({}) + {} * ({})'.format(g,a,u,b,v))

# Les entiers inversibles modulo n

def invmod(a,n):
    '''Retourne l'inverse de a modulo n s'il existe.'''
    (g,u,v) = bezout(a,n)
    if g != 1 : return None      # mieux que de provoquer une erreur...
    # sinon g = 1 = a*u + n*v = a*u (mod n) donc l'inverse de a modulo n est u
    return u % n

if __name__ == '__main__':
    from random import randint
    a = randint(2,1000000000000000000000000000000000000)
    b = randint(2,1000000000000000000000000000000000000)
    print('pgcd({},{}) == {}'.format(a,b,pgcd(a,b)))
    print('{} et {} sont-ils premiers entre eux ?'.format(a,b),end=' ')
    print('Oui' if premiers_entre_eux(a,b) else 'Non')
    print()
    p = proba_premiers_entre_eux()
    print('Proba expérimentale que deux entiers soient premiers entre eux :',p)
    from math import pi
    print('Proba théorique : 6/pi**2 = ',6/pi**2)
    print()
    print('Décomposition de Bezout du pgcd de 8,30 puis de 8,11 :')
    aff_bezout(8,30)
    aff_bezout(8,11)
    print()
    print("L'inverse de 5 modulo 9 est",invmod(5,9))     # 2
    essai = invmod(6,8)     # est-il inversible ?
    if essai:
        print("L'inverse de 6 modulo 8 est", essai)     # oups !
    else:
        print("6 n'est pas inversible modulo 8")
    
    # COMPLEMENTS EN LIGNE
    from webbrowser import open as browse
    #browse('https://fr.wikipedia.org/wiki/Nombres_premiers_entre_eux')
    #browse('https://fr.wikipedia.org/wiki/Théorème_de_Bachet-Bézout')
    #browse('http://gjmaths.pagesperso-orange.fr/contenu/lamenon.pdf')
    
    

    
    


    





      
