#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
### PYTHON. APPRENTISSAGE ACTIF (JP. ROY, EDITIONS ELLIPSES, 2020) ###
######################################################################

"""
exo_puisdicho.py (11.19)
Calcul d'une puissance x**n par dichotomie.
"""

# L'algorithme naïf fait un nombre LINEAIRE de multiplications. Coût en O(n).

def puis_rec(x,n):  # x * x * x * ... * x par récurrence sur n
    if n == 0: return 1
    return x * puis(x,n - 1)
    
def puis(x,n):      # x * x * x * ... * x
    res = 1         # l'élément neutre de la multiplication
    for i in range(n):
        res = res * x
    return res

# Accélération d'algorithme par DICHOTOMIE (division par 2) :
# x**0 = 1                                         2**0  = 1
# x**n = (x*x)**(n//2)       si n pair             2**10 = (2*2)**5
# x**n = x * (x*x)**(n//2)   si n impair           2**11 = 2 * (2*2)**5 

# D'abord par RECURRENCE, souvent le plus simple à concevoir.
    
def puisdicho_rec(x,n):               # récurrence sur n
    '''Calcule par récurrence et dichotomie x**n avec n entier >= 0.'''
    if n == 0: return 1
    if n % 2 == 0: return puisdicho_rec(x*x,n // 2)
    return x * puisdicho_rec(x*x,n // 2)

# Le nombre de pas h de l'algorithme ci-dessus est en O(log n). Pourquoi ? 
# On fait grosso modo h divisions par 2 avant que n tombe sur 0. Donc 
# inversement 2**h est approximativement égal à n, donc h est le logarithme 
# en base 2 de n. Puisque tous les logarithmes sont proportionnels entre
# eux, le nombre de pas est donc LOGARITHMIQUE, en O(log n). Très rapide !

# Produisons enfin un algorithme ITERATIF. Le raisonnement est un peu plus
# fin. Le code de puisdicho_rec en dernière ligne calcule une expression de 
# la forme a * b**k. Introduisons donc la fonction auxilliaire :
#                            aux(a,b,k) = a * b**k
# et raisonnons sur elle à partir de la formulation par récurrence. 
# Notons que :
#                        puisdicho(x,n) = aux(1,x,n)
# - si k = 0 : 
#     aux(a,b,k) = a
# - si k est pair : 
#     aux(a,b,k) = a * b**k = a * (b*b)**(k//2) = aux(a,b*b,k//2)
# - si k est impair : 
#     aux(a,b,k) = a * b**k = a * b * (b*b)**(k//2) = aux(a*b,b*b,k//2)
# d'où une nouvelle version dans laquelle la récurrence est en tête (on dit
# aussi "terminale", en anglais "tail recursion"), que l'on transformera 
# ensuite facilement en boucle ! Certains langages (Scheme, Javascript, CAML,
# voire certaines versions de C comme gcc) savent le faire de manière
# automatique, au moins dans les cas simples comme ici. D'autres ne le font
# pas, comme Java ou Python, obligeant à produire une boucle.

def aux(a, b, k):    # récurrence terminale mais consomme de la pile en Python
    #print('a =',a,' b =',b,'k =',k)
    if k == 0: return a
    if k % 2 == 0: return aux(a, b*b, k//2)
    return aux(a*b, b*b, k//2)

def puisdicho_rec_term(x,n):
    return aux(1, x, n)
        
# Transformation quasi-automatique de aux en une véritable boucle while, 
# ce qui était trop difficile à partir de la "vraie" récurrence de la 
# première version...

def puisdicho_iter(x, n) :     # n entier >= 0, version finale !
    (a, b, n) = (1, x, n)
    while n > 0:
        if n % 2 == 0: (a, b, n) = (a, b*b, n//2)
        else: (a, b, n) = (a*b, b*b, n//2)
    return a

if __name__ == '__main__' :
    import sys ; sys.setrecursionlimit(5000)   # pour la récurrence...
    print("2**10 avec l'algorithme naïf (récursif) :",puis(2,10))
    print("2**10 avec l'algorithme naïf (itératif) :",puis(2,10))
    print()
    print('Algorithme dichotomique :')
    print('- Récurrence ("profonde" ou "enveloppée", non terminale)')
    print('2**10 =',puisdicho_rec(2,10))
    print('2**11 =',puisdicho_rec(2,11))
    print('- Récurrence "terminale"')
    print(' (fausse itération en Python et Java)')
    print(' (vraie itération en Scheme, CAML, JavaScript voire en C optimisé)')
    print('2**10 =',puisdicho_rec_term(2,10))
    print('2**11 =',puisdicho_rec_term(2,11))
    print('- Itération (boucle)')
    print('2**10 =',puisdicho_iter(2,10))
    print('2**11 =',puisdicho_iter(2,11))
    
    # COMPLEMENTS EN LIGNE
    from webbrowser import open as browse
    #browse('https://fr.wikipedia.org/wiki/Récursion_terminale')

