#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
### PYTHON. APPRENTISSAGE ACTIF (JP. ROY, EDITIONS ELLIPSES, 2020) ###
######################################################################

'''
exo_retropropagation.py
L'algorithme de rétro-propagation du gradient dans un
réseau de neurones avec une couche cachée.
'''

from random import uniform
from math import exp, pow
from exo_mat2 import Mat
from time import time

def sig(x):    # la fonction sigmoïde, dérivable
    return 1/(1+ exp(-x))

class ReseauRetroPropagation():
    def __init__(self,ne=2,nc=3,ns=1,nbiter=3,eta=1):
        '''Construit un réseau de neurones avec une couche cachée. Il y a ne entrées (+ biais),
        nc neurones dans la couche cachée (+ biais) et ns neurones en sortie.'''
        self.ne = ne
        self.nc = nc
        self.ns = ns
        print(self.ne,'entrées(+1),',self.nc,'neurones cachés(+1) et',self.ns,'en sortie.')
        # le réseau calcule sur 7 vecteurs et 2 matrices
        self.act_i = [1] + ne * [0]     # les entrées concrètes seront fournies avec la méthode accepte
        self.mat_ij = Mat(lambda j,i: uniform(-1,1),nc,ne+1)  # self.mat_ij[j][i] == poids i->j
        self.stim_j = [1] + nc * [0]    # les stimuli \tilde{z}_j reçus par les neurones cachés
        self.act_j = [1] + nc * [0]     # valeurs z_j des neurones cachés
        self.grad_j = [0] + nc * [0]    # gradients locaux des neurones cachés
        self.mat_jk = Mat(lambda k,j: uniform(-1,1),ns,nc+1)  # self.mat_jk[k][j] == poids j->k
        self.stim_k = ns * [0]          # les stimuli \tilde{z}_j reçus par les neurones cachés
        self.act_k = ns * [0]           # valeurs z_k des neurones de sortie
        self.grad_k = ns * [0]          # gradients locaux des neurones de sortie
        self.nbiter = nbiter
        self.eta = eta                  # "learning rate"
        self.error = 0

    def accepte(self,Lentrees):         # on entre des entrées et on les propage
        if len(Lentrees) != len(self.act_i) - 1:
            raise ValueError("Mauvais nombre d'entrées !")
        self.act_i[1:] = Lentrees       # on ne touche pas au biais
        self.propage()                  # propagation des entrées
        return self.act_k               # et retour des sorties

    def propage(self) :                # propagation des entrées vers la sortie
        # calcul des stimuli reçus par la couche cachée à partir des entrées
        self.stim_j = self.mat_ij.app(self.act_i)
        # calcul des réponses des neurones cachés
        self.act_j = [1] + list(map(sig,self.stim_j))
        # calcul des stimuli reçus par la couche de sortie
        self.stim_k = self.mat_jk.app(self.act_j)
        # calcul des réponses des neurones cachés
        self.act_k = list(map(sig,self.stim_k))

    def apprentissage(self,Lexemples):  # apprentissage des poids par une liste d'exemples
        ip = 0                          # numéro de l'exemple courant
        for it in range(self.nbiter):   # le nombre d'itérations est fixé !
            error = 0.0                 # l'erreur totale pour cet exemple
            (entrees,sorties_attendues) = Lexemples[ip]         # un nouvel exemple à apprendre
            # PROPAGATION VERS L'AVANT
            self.accepte(entrees)       # sorties obtenues sur l'exemple courant, self.act_k est mis à jour
            # RETRO_PROPAGATION VERS L'ARRIERE, EN DEUX TEMPS
            # TEMPS 1. calcul des gradients locaux sur la couche k de sortie (les erreurs commises)
            for k in range(self.ns):
                self.grad_k[k] = sorties_attendues[k] - self.act_k[k]       # gradient sur un neurone de sortie (erreur locale)
                error += pow(self.grad_k[k],2)                              # l'erreur quadratique totale
            error *= 0.5
            if it == self.nbiter-1 : self.error = error                     # mémorisation de l'erreur totale à la dernière itération
            # modification des poids j->k
            for j in range(self.nc+1):
                for k in range(self.ns):
                    self.mat_jk.L[k][j] -= -self.eta*self.act_j[j]*self.act_k[k]*(1-self.act_k[k])*self.grad_k[k]
            # Réponse à  la question "b4" : T_{jk} = z_k * (1-z_k) * w_{jk}
            # TEMPS 2. calcul des gradients locaux sur la couche j cachée (rétro-propagation), sauf pour le bias constant
            for j in range(1,self.nc+1):
                self.grad_j[j] = sum(self.act_k[k]*(1-self.act_k[k])*self.mat_jk.L[k][j]*self.grad_k[k] for k in range(self.ns))
            # modification des poids i->j
            for i in range(self.ne+1):
                for j in range(1,self.nc):
                    self.mat_ij.L[j-1][i] -= -self.eta*self.act_i[i]*self.act_j[j]*(1-self.act_j[j])*self.grad_j[j]
            # et l'on passe à l'exemple suivant
            if trace and (it in (0,self.nbiter-1)) and (ip == len(Lexemples)-1):
                self.dump(it,'sortie')
            ip = (ip + 1) % len(Lexemples)      # parcours des exemples en ordre circulaire

    def dump(self,n,msg):     # dump du réseau en entrant dans l'itération numéro n
        print('---------- DUMP',msg,'itération numéro',n)
        print('mat_ij :') ; print(self.mat_ij)
        print('act_j  :',self.act_j)
        print('grad_j :',self.grad_j)
        print('mat_jk :') ; print(self.mat_jk)
        print('act_k  :',self.act_k)
        print('grad_k :',self.grad_k)
        print()

    def test(self,Lexemples):
        print('Test des exemples :')
        for (entree,sortie_attendue) in Lexemples:
            self.accepte(entree)
            print(entree,'-->',self.act_k,': on attendait',sortie_attendue)

if __name__ == '__main__':
    trace = False
    print('################## NOT ##################')
    r1 = ReseauRetroPropagation(1,2,1,nbiter=10000,eta=0.5)
    Lexemples1 = [[[1],[0]],[[0],[1]]]
    START = time() ; r1.apprentissage(Lexemples1) ; END = time()
    r1.test(Lexemples1)
    print('APPRENTISSAGE sur {} itérations, time = {:.2f} sec'.format(r1.nbiter,END-START))
    print()
    print('################## XOR ##################')
    r2 = ReseauRetroPropagation(2,3,1,nbiter=50000,eta=0.1)
    Lexemples2 = [[[1,0],[1]], [[0,0],[0]], [[0,1],[1]], [[1,1],[0]]]
    START = time() ; r2.apprentissage(Lexemples2) ; END = time()
    print('APPRENTISSAGE sur {} itérations, time = {:.2f} sec'.format(r2.nbiter,END-START))
    r2.test(Lexemples2)

    # COMPLEMENTS EN LIGNE (qui débordent parfois du simple perceptron)
    from webbrowser import open as browse
    #browse('https://en.wikipedia.org/wiki/Perceptron')      # meilleur que la version française
    #browse('https://www.youtube.com/watch?v=5w-jolCle6g')   # 5 minutes :-)
    
    #browse('http://www.cs.toronto.edu/~hinton/') # cherchez le mot COURSERA et regardez les vidéos !
    # Le MOOC de Geoffrey Hinton (un pionnier) de 2012 n'est plus en activité. Pour quelque chose
    # de plus récent, il y a (parmi d'autres) celui d'Andrew Ng sur Coursera :
    # https://www.coursera.org/learn/neural-networks-deep-learning
    
    #browse('https://www.irif.fr/~kesner/enseignement/iup/cours81.pdf')
    #browse('https://en.wikipedia.org/wiki/History_of_artificial_intelligence#Perceptrons_and_the_dark_age_of_connectionism')

     #browse('https://epynn.net')
     




