#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
### PYTHON. APPRENTISSAGE ACTIF (JP. ROY, EDITIONS ELLIPSES, 2020) ###
######################################################################

"""
exo_yacc1.py (11.113)
Illustration de SLY, logiciel Lex/Yacc pour Python de D. Beazley.
Analyseur syntaxique pour des expressions arithmétiques parenthésées
sur les opérateurs + - * / ** avec des nombres entiers ou flottants
et des variables.
Ce fichier est la SOLUTION de l'exercice 11.113 du livre. L'exercice
consistait à compléter le fichier "exo_yacc1_student.py"
"""

from exo_lex1 import ArithLexer, LexerError         # notre analyseur lexical
from sly import Parser                 # la classe Parser à sous-classer

class ParserError(Exception):          # juste pour la lisibilité des exceptions
    pass

class ArithParser(Parser):             # une sous-classe de Parser
    tokens = ArithLexer.tokens         # les noms symboliques des tokens
    # {VAR, FLOAT, INT, ADD, SUB, PUIS, MUL, DIV, PARG, PARD}

    def __init__(self,trace=False):    # trace pour voir les règles appliquées
        Parser.__init__(self)
        self.trace = trace

    # Grammaire des expressions
    #    (R1)  expr ::= VAR
    #    (R2)  expr ::= FLOAT
    #    (R3)  expr ::= INT
    #    (R4)  expr ::= expr ADD expr   |
    #    (R4)  expr ::= expr SUB expr   |
    #    (R4)  expr ::= expr MUL expr   |  regroupées dans une même méthode
    #    (R4)  expr ::= expr DIV expr   |
    #    (R4)  expr ::= expr PUIS expr  |
    #    (R5)  expr ::= PARG expr PARD
    #    (R6)  expr ::= SUB expr %prec USUB        # le moins unaire

    # Cette grammaire est ambiguë, il faut rajouter des priorités pour
    # permettre au parser de lever l'ambiguïté. Régles de "précédence" :

    precedence = [
        # + et - ont mêmes priorités, et associatives à gauche
        ('left','ADD','SUB'),       # x+y+z == (x+y)+z, idem avec -
        # idem pour * et / mais qui sont plus prioritaires que + et -
        ('left','MUL','DIV'),       # x*y*z == (x*y)*z, idem avec /
        # opérateur virtuel pour l'opérateur "- unaire"
        ('right','USUB'),
        # ** est la plus prioritaire, et associative à droite
        ('right','PUIS')
    ]

    # Les règles de grammaire proprement dites :

    @_('VAR')
    def expr(self,p):
        if self.trace: print('Règle R1 :',p[0])
        return p[0]

    @_('INT')
    def expr(self,p):
        if self.trace: print('Règle R2 :',p[0])
        return p[0]                  # un entier

    @_('FLOAT')
    def expr(self,p):
        if self.trace: print('Règle R3 :',p[0])
        return p[0]                  # un flottant

    @_('expr ADD expr',
       'expr SUB expr',
       'expr MUL expr',
       'expr DIV expr',
       'expr PUIS expr')  # même action !
    def expr(self,p):
        if self.trace: print('Règle R4/{1} : {0}, {1}, {2}'.format(p[0],p[1],p[2]))
        return [p[1],p[0],p[2]]   # traduction vers du Scheme-like (un arbre d'analyse)

    @_('PARG expr PARD')
    def expr(self,p):
        if self.trace: print('Règle R5 :',p[0],p[1],p[2])
        return p[1]                  # on élimine les parenthèses

    @_('expr : SUB expr %prec USUB')
    def expr(self,p):
        if self.trace: print('Règle R6 :',p[0],p[1])
        return ['*',-1,p[1]]         # le choix de sympy. Ou alors arbre('-',0,p[1])

    def error(self,t):   # gestion des erreurs plus complète dans la doc de SLY...
        if t: raise ParserError('sur ' + str(t))
        raise ParserError('expression incomplète')


if __name__ == '__main__':
    print('Proposez des expressions arithmétiques sur les opérateurs')
    print('+ - * / ** sur entiers et flottants, avec variables.')
    print('Entrez .fini pour quitter. Un exemple :')
    lexer = ArithLexer()                # l'analyseur lexical
    parser = ArithParser(trace=True)    # l'analyseur syntaxique (réglez trace)

    def parser_pre(s):              # résultat en préfixe (style Lisp/Scheme)
        tok = lexer.tokenize(s)     # le générateur de tokens à la demande
        return parser.parse(tok)    # arbre d'analyse syntaxique du flot de tokens

    def postfixe(p):           # liste postfixe plate du résultat du parser
        if type(p) != list: return [p]
        return postfixe(p[1]) + postfixe(p[2]) + [p[0]]

    def infixe(p):             # liste infixe complètement parenthésée
        if type(p) != list: return p
        return [infixe(p[1]), p[0], infixe(p[2])]

    s = '-x2*(1.5+y**2**3)-z'       # un exemple d'expression...
    print()
    print('Conversion de',repr(s),':')
    res = parser_pre(s)
    print('Préfixe parenthésé :\n',res)
    print('Postfixe plat :\n',postfixe(res))
    print('Infixe parenthésé  :\n',infixe(res))
    print()
    parser.trace = False
    while True:                         # .fini pour quitter proprement
        try:
            s = input('votre expression : ')
            if s.lower() == '.fini' : break
            res = parser_pre(s)
            print('Préfixe :\n',res)
            print('Postfixe plat :\n',postfixe(res))
            print('Infixe :\n',infixe(res))
        except LexerError as e:
            print("LexerError " + str(e))
        except ParserError as e:
            print("ParserError " + str(e))

    from webbrowser import open as browse
    #browse('https://www.youtube.com/watch?v=zJ9z6Ge-vXs')       # SLY @ PYCON 2018
    #browse('https://sly.readthedocs.io/en/latest/index.html')   # doc de SLY



