#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
### PYTHON. APPRENTISSAGE ACTIF (JP. ROY, EDITIONS ELLIPSES, 2020) ###
######################################################################

"""
code_yacc.py (§8.3)
Illustration de SLY, logiciel Lex/Yacc pour Python de Dave Beazley.
Analyseur syntaxique pour des expressions arithmétiques parenthésées
sur les opérateurs + - * / et avec des nombres entiers positifs seulement.
On ne gère pas le - unaire.
"""

from code_lex import ArithLexer, LexerError   # notre lexer d'expressions arithmétiques
from sly import Parser                 # la classe Parser à sous-classer
import operator                        # pour les *fonctions* + - * /

class ParserError(Exception):          # juste pour la lisibilité des exceptions
    pass

class ArithParser(Parser):             # une sous-classe de Parser
    tokens = ArithLexer.tokens         # en premier lieu, les tokens du lexer

    # Grammaire des expressions, en utilisant les noms symboliques des tokens
    #    (R0)  initial ::= expr        # je veux une expression bien formée
    #    (R1)  expr ::= INT            # <--- positifs seulement dans ce fichier
    #    (R2)  expr ::= PARG expr PARD
    #    (R3)  expr ::= expr ADD expr
    #    (R4)  expr ::= expr SUB expr
    #    (R5)  expr ::= expr MUL expr
    #    (R6)  expr ::= expr DIV expr

    # Cette grammaire est ambiguë, il faut rajouter des priorités pour
    # permettre au parser de lever l'ambiguïté. Régles de "précédence" :

    precedence = [
        # + et - ont mêmes priorités, et associatives à gauche
        ('left','ADD','SUB'),       # x+y+z == (x+y)+z, idem avec -
        # idem pour * et / mais qui sont plus prioritaires que + et -
        ('left','MUL','DIV'),       # x*y*z == (x*y)*z, idem avec /
    ]

    # Les règles de grammaire proprement dites :

    @_('expr')
    def initial(self,p):
        return p[0]

    @_('INT')
    def expr(self,p):
        print('Règle R1 :',p[0])
        return p[0]                  # un entier

    @_('PARG expr PARD')
    def expr(self,p):
        print('Règle R2 :',p[0],p[1],p[2])
        return p[1]                  # on élimine les parenthèses

    @_('expr ADD expr', 'expr SUB expr', 'expr MUL expr', 'expr DIV expr')  # même action !
    def expr(self,p):
        print('Règle',p[1],':',p[0],p[1],p[2])
        return [p[1], p[0], p[2]]   # traduction vers du Scheme-like (un arbre d'analyse)
        # ou passage d'un traducteur à une calculette :
        # return ArithParser.operateur(p[1])(p[0],p[2])           #     (*)

    # utilitaire pour jouer avec la ligne (*)
    @staticmethod
    def operateur(op):
        if op == '+' : return operator.__add__
        elif op == '-' : return operator.__sub__
        elif op == '*' : return operator.__mul__
        elif op == '/' : return operator.__truediv__
        else : raise ValueError('Opérateur inconnu : ' + op)

    def error(self,t):   # gestion des erreurs plus complète dans la doc de SLY...
        if t:
            raise ParserError('sur ' + str(t))
        else:
            raise ParserError('expression incomplète')


if __name__ == '__main__':
    print('Proposez des expressions arithmétiques sur les opérateurs')
    print('+ - * / et ne comportant que des entiers positifs.')
    print('Entrez .fini pour quitter.')
    lexer = ArithLexer()                # l'analyseur lexical
    parser = ArithParser()              # l'analyseur syntaxique
    while True:                         # .fini pour quitter proprement
        try:
            s = input('votre expression : ')
            if s.lower() == '.fini' : break
            tok = lexer.tokenize(s)     # le générateur de tokens à la demande
            res = parser.parse(tok)     # analyse syntaxique du flot de tokens
            print('Conversion de',repr(s),':')
            print(res)
        except LexerError as e:
            print("LexerError " + str(e))
        except ParserError as e:
            print("ParserError " + str(e))

    from webbrowser import open as browse
    #browse('http://epaperpress.com/lexandyacc/thy.html')
    #browse('https://www.youtube.com/watch?v=zJ9z6Ge-vXs')       # SLY @ PYCON 2018
    #browse('https://sly.readthedocs.io/en/latest/index.html')   # doc de PLY




